@@ -1,12 +1,13 @@
-
 from .season import Season
 from .meeting import Meeting
 from .session import Session
 from .driver import Driver
+from .circuit import Circuit
 
 __all__ = [
     "Season",
     "Meeting",
     "Session",
-    "Driver"
+    "Driver",
+    "Circuit"
 ]
\ No newline at end of file
