@@ -157,7 +157,8 @@ class Season:
             return self.__str__() # Print the meetings table.
         else:
             display(self.season_table) # Display the meetings table.
-            return self.__str__()
+            # return self.__str__()
+            return ""
 
     def __str__(self):
         """
