@@ -2,6 +2,7 @@
 from urllib.parse import urljoin
 from typing import List, Dict
 from time import time
+import functools
 
 # Third-Party Library Imports
 # (No third-party libraries imported in this file)
@@ -12,7 +13,9 @@ from ..utils import helper
 from ..utils.logger import logger
 from ..data_processing.etl import *
 from ..data_processing.data_models import *
+from ..data_processing.silver_functions import *
 from ..utils.constants import TOPICS_MAP, SILVER_SESSION_TABLES, TABLE_GENERATION_FUNCTIONS
+from ..utils.exceptions import *
 from ..data_processing.lakes import DataLake
 from .driver import Driver
 
@@ -86,7 +89,17 @@ class Session:
         # Build the full path for accessing session data if path attribute exists.
         if hasattr(self, "path"):
             self.full_path = helper.build_session_endpoint(self.path)
+
+    def _load_default_silver_tables(self):
+        for table_name in SILVER_SESSION_TABLES:
+            self.create_silver_table(table_name, TABLE_REQUIREMENTS[table_name], include_session=True)(globals()[TABLE_GENERATION_FUNCTIONS[table_name]])
     
+
+    def _load_circuit_data(self):
+        circuit = self.meeting.circuit
+        circuit._load_circuit_data()
+        self.data_lake.create_bronze_table(table_name="track_regions", raw_data=circuit._raw_circuit_data, parsed_data=circuit.track_regions)
+
     def load_session_data(self):
         """
         Load the session data.
@@ -205,21 +218,29 @@ class Session:
             self.drivers[driver.RacingNumber] = driver
 
 
-    def get_driver(self, driver_number: str) -> Driver:
+    def get_driver(self, identifier: str) -> Driver:
         """
-        Get a specific driver by their number.
+        Get a specific driver by their number, name, or short name.
 
         Parameters
         ----------
-        driver_number : str
-            The driver's racing number
+        identifier : str
+            The driver's racing number, full name, or short name.
 
         Returns
         -------
         Driver
-            The Driver object for the specified number
+            The Driver object for the specified identifier, or None if not found.
         """
-        return self.drivers.get(str(driver_number))
+        for driver in self.drivers.values():
+            if (
+                str(driver.RacingNumber) == identifier or
+                driver.FirstName.lower() == identifier.lower() or
+                driver.LastName.lower() == identifier.lower() or
+                driver.Tla.lower() == identifier.lower()
+            ):
+                return driver
+        return None
 
     def load_data(
         self,
@@ -259,25 +280,13 @@ class Session:
 
         # Handle single data name case
         single_input = len(dataNames) == 1
-        # if single_input:
-        #     dataNames = [dataNames]
-
-        # # Validate all data names
-        # validated_names = []
-        # for name in dataNames:
-        #     for topic in self.topic_names_info:
-        #         if self.topic_names_info[topic]["key"] == name:
-        #             name = topic
-        #             stream = self.topic_names_info[topic]["default_is_stream"]
-        #             break
-        #     validated_names.append((name, stream))
-
         validated_names = dataNames
 
         results = {}
         if parallel and len(validated_names) > 1:
             # Parallel loading
             n_processes = max(1, multiprocessing.cpu_count() - 1)
+            print(validated_names)
             with Pool(processes=n_processes) as pool:
                 loaded_results = pool.starmap(
                     load_single_data, 
@@ -285,43 +294,76 @@ class Session:
                         np.asarray(validated_names)[:,0],
                         repeat(self),
                         np.asarray(validated_names)[:,1]))
-                results = {name: result for name, result in loaded_results}
+                results = {name: {"data": data, "parsed_data": parsed_data} for name, data, parsed_data in loaded_results}
         else:
             # Sequential loading
             for name, stream in validated_names:
-                name, res = load_single_data(name, self, stream)
-                # logger.info(f"Fetching data : '{name}'")
-                # start = time()
-                # data = livetimingF1_getdata(
-                #     urljoin(self.full_path, self.topic_names_info[name][dataType]),
-                #     stream=stream
-                # )
-                # logger.debug(f"Fetched in {round(time() - start,3)} seconds")
-                
-                # start = time()
-                # res = BasicResult(
-                #     data=list(self.etl_parser.unified_parse(name, data))
-                # )
-                # logger.debug(f"Parsed in {round(time() - start,3)} seconds")
-                # logger.info(f"'{name}' has been fetched and parsed")
-                results[name] = res
+                name, data, parsed_data = load_single_data(name, self, stream)
+                results[name] = {"data": data, "parsed_data": parsed_data}
 
         # Save all results to bronze lake
         for name, result in results.items():
-            self.data_lake.put(
-                level="bronze", data_name=name, data=result
-            )
+            self.data_lake.create_bronze_table(table_name=name, raw_data=result["data"], parsed_data=result["parsed_data"])
             logger.debug(f"'{name}' has been saved to the bronze lake.")
 
         # Return single result or dict based on input type
         if single_input:
-            return self.data_lake.get(level="bronze", data_name=validated_names[0][0])
-        return {name: self.data_lake.get(level="bronze", data_name=name) 
+            return self.data_lake.get(level="bronze", table_name=validated_names[0][0])
+        return {name: self.data_lake.get(level="bronze", table_name=name)
                for name, stream in validated_names}
 
+    def get_table(
+        self,
+        dataNames,
+        level = "bronze",
+        parallel: bool = False,
+        force: bool = False
+    ):
+        if isinstance(dataNames, str):
+            return self.get(
+                dataNames,
+                level=level,
+                parallel= parallel,
+                force = force)
+
+        elif isinstance(dataNames, list):
+            res = self.get(
+                dataNames,
+                level=level,
+                parallel= parallel,
+                force = force
+                )
+            return {name:table for name, table in res.items()}
+        else: return None
+
     def get_data(
         self,
         dataNames,
+        level = "bronze",
+        parallel: bool = False,
+        force: bool = False
+    ):
+        if isinstance(dataNames, str):
+            return self.get(
+                dataNames,
+                level=level,
+                parallel= parallel,
+                force = force).df
+
+        elif isinstance(dataNames, list):
+            res = self.get(
+                dataNames,
+                level=level,
+                parallel= parallel,
+                force = force
+                )
+            return {name:table.df for name, table in res.items()}
+        else: return None
+
+    def get(
+        self,
+        dataNames,
+        level = "bronze",
         parallel: bool = False,
         force: bool = False
     ):
@@ -367,44 +409,65 @@ class Session:
         - Uses parallel processing for multiple topics when parallel=True
         - Returns same format as input: single result for str input, dict for list input
         """
-        # Ensure topic names are loaded
-        if not hasattr(self, "topic_names_info"):
-            self.get_topic_names()
-        
-        # Handle single data name case
-        single_input = isinstance(dataNames, str)
-        dataNames = [dataNames] if single_input else dataNames
-        
-        # Validate all data names
-        validated_names = [self.check_data_name(name) for name in dataNames]
-        
-        # Check cache and identify topics to load
-        to_load = []
-        results = {}
-        
-        for name in validated_names:
-            if not force and name in self.data_lake.raw:
-                logger.debug(f"'{name}' found in lake, using cached version")
-                results[name] = self.data_lake.get(level="bronze", data_name=name)
-            else:
-                stream = self.topic_names_info[name]["default_is_stream"]
-                to_load.append((name, stream))
-        
-        if to_load:
-            # Load new data using load_data with parallel option
-            loaded_results = self.load_data(
-                dataNames=to_load,
-                parallel=parallel and len(to_load) > 1
-            )
+
+        if level == "bronze":
+            # Ensure topic names are loaded
+            if not hasattr(self, "topic_names_info"):
+                self.get_topic_names()
             
-            if isinstance(loaded_results, dict):
-                results.update(loaded_results)
-            else:
-                # Handle single result case
-                results[to_load[0][0]] = loaded_results
+            # Handle single data name case
+            single_input = isinstance(dataNames, str)
+            dataNames = [dataNames] if single_input else dataNames
+            
+            # Validate all data names
+            validated_names = [self.check_data_name(name) for name in dataNames]
+            
+            # Check cache and identify topics to load
+            to_load = []
+            results = {}
+            
+            for name in validated_names:
+                if not force and name in self.data_lake.metadata:
+                    logger.debug(f"'{name}' found in lake, using cached version")
+                    results[name] = self.data_lake.get(level="bronze", table_name=name)
+                else:
+                    logger.debug(f"'{name}' not found in lake, loading from livetiming.")
+                    stream = self.topic_names_info[name]["default_is_stream"]
+                    to_load.append((name, stream))
+            
+            if to_load:
+                # Load new data using load_data with parallel option
+                loaded_results = self.load_data(
+                    dataNames=to_load,
+                    parallel=parallel and len(to_load) > 1
+                )
+                
+                if isinstance(loaded_results, dict):
+                    results.update(loaded_results)
+                else:
+                    # Handle single result case
+                    results[to_load[0][0]] = loaded_results
+            
+            # Return single result if single input, otherwise return dictionary
+            return results[validated_names[0]] if single_input else results
         
-        # Return single result if single input, otherwise return dictionary
-        return results[validated_names[0]] if single_input else results
+        else:
+            # Handle single data name case
+            single_input = isinstance(dataNames, str)
+            dataNames = [dataNames] if single_input else dataNames
+
+            results = {}
+            for name in dataNames:
+                if name in self.data_lake.metadata:
+                    if self.data_lake.metadata[name]["generated"]:
+                        logger.debug(f"'{name}' has been created and generated.")
+                        results[name] = self.data_lake.get(level=level, table_name=name)
+                    else:
+                        logger.debug(f"'{name}' has been created but not generated yet. It is being generated...")
+                        self.data_lake.get(level=level, table_name=name).generate_table()
+                        results[name] = self.data_lake.get(level=level, table_name=name)
+            
+            return results[dataNames[0]] if single_input else results
 
     def check_data_name(self, dataName: str):
         """
@@ -432,12 +495,39 @@ class Session:
             self.get_topic_names()
 
         for topic in self.topic_names_info:
-            if self.topic_names_info[topic]["key"] == dataName:
+            if (self.topic_names_info[topic]["key"].lower() == dataName.lower()) or (topic.lower() == dataName.lower()):
                 dataName = topic
-                break
-
+                return dataName
+        
         return dataName
 
+        # raise TopicNotFoundError(f"The topic name you provided '{dataName}' is not included in the topics.")
+    
+    # def normalize_topic_name(self, topicName):
+    #     """
+    #     Normalize the topic name.
+    #     """
+
+    #     if not hasattr(self,"topic_names_info"):
+    #         self.get_topic_names()
+
+    #     for topic in self.topic_names_info:
+    #         if (self.topic_names_info[topic]["key"].lower() == topicName.lower()) or (topic.lower() == topicName.lower()):
+    #             topicName = topic
+    #             break
+
+    #     return topicName
+
+    def _identify_data_level(self, dataName):
+
+        corrected_data_name = self.check_data_name(dataName)
+        if corrected_data_name: return "bronze"
+        else:
+            for table_name, info in self.data_lake.metadata.items():
+                if dataName == table_name: return info["table_type"]
+        
+        return None
+
     def get_laps(self):
         """
         Retrieve the laps data.
@@ -484,63 +574,55 @@ class Session:
             logger.info("Car Telemetry table is not generated yet. Use .generate() to load required data and generate silver tables.")
             return None
 
-    def get_weather(self):
-        """
-        Retrieve the weather data.
+    # def get_weather(self):
+    #     """
+    #     Retrieve the weather data.
 
-        This method returns the weather data if it has been generated. If not, it logs an 
-        informational message indicating that the weather table is not generated yet.
+    #     This method returns the weather data if it has been generated. If not, it logs an 
+    #     informational message indicating that the weather table is not generated yet.
 
-        Returns
-        -------
-        :class:`~Weather` or None
-            The weather data if available, otherwise None.
+    #     Returns
+    #     -------
+    #     :class:`~Weather` or None
+    #         The weather data if available, otherwise None.
 
-        Notes
-        -----
-        - The method checks if the `weather` attribute is populated.
-        - If the `weather` attribute is not populated, it logs an informational message.
-        """
+    #     Notes
+    #     -----
+    #     - The method checks if the `weather` attribute is populated.
+    #     - If the `weather` attribute is not populated, it logs an informational message.
+    #     """
 
-        logger.error(".get_weather() is not implemented yet.")
-
-        # if self.weather != None: return self.weather
-        # else:
-        #     logger.info("Weather table is not generated yet. Use .generate() to load required data and generate silver tables.")
-        #     return None
+    #     logger.error(".get_weather() is not implemented yet.")
     
-    def get_timing(self):
-        """
-        Retrieve the timing data.
+    # def get_timing(self):
+    #     """
+    #     Retrieve the timing data.
 
-        This method returns the timing data if it has been generated. If not, it logs an 
-        informational message indicating that the timing table is not generated yet.
+    #     This method returns the timing data if it has been generated. If not, it logs an 
+    #     informational message indicating that the timing table is not generated yet.
 
-        Returns
-        -------
-        :class:`~Timing` or None
-            The timing data if available, otherwise None.
+    #     Returns
+    #     -------
+    #     :class:`~Timing` or None
+    #         The timing data if available, otherwise None.
 
-        Notes
-        -----
-        - The method checks if the `timing` attribute is populated.
-        - If the `timing` attribute is not populated, it logs an informational message.
-        """
+    #     Notes
+    #     -----
+    #     - The method checks if the `timing` attribute is populated.
+    #     - If the `timing` attribute is not populated, it logs an informational message.
+    #     """
 
-        logger.error(".get_timing() is not implemented yet.")
-
-        # if self.timing != None: return self.timing
-        # else:
-        #     logger.info("Timing table is not generated yet. Use .generate() to load required data and generate silver tables.")
-        #     return None
+    #     logger.error(".get_timing() is not implemented yet.")
     
     def _get_first_datetime(self):
         pos_df = self.get_data("Position.z")
         car_df = self.get_data("CarData.z")
-        first_date = np.amax([(helper.to_datetime(car_df["Utc"]) - pd.to_timedelta(car_df["timestamp"])).max(), (helper.to_datetime(pos_df["Utc"]) - pd.to_timedelta(pos_df["timestamp"])).max()])
-        
-        # sess_data = self.get_data("Session_Data")
-        # first_date = helper.to_datetime(sess_data[sess_data["SessionStatus"] == "Started"].Utc).tolist()[0]
+        first_date = np.amax(
+            [
+                (helper.to_datetime(car_df["Utc"]) - pd.to_timedelta(car_df["timestamp"])).max(),
+                (helper.to_datetime(pos_df["Utc"]) - pd.to_timedelta(pos_df["timestamp"])).max()
+            ]
+        )
         return first_date
     
     def _get_session_start_datetime(self):
@@ -550,46 +632,186 @@ class Session:
         return first_date
 
     def generate(self, silver=True, gold=False):
+        
+        self._load_circuit_data()
+
         required_data = set(["CarData.z", "Position.z", "SessionStatus"])
         tables_to_generate = set()
         if silver:
-            tables_to_generate.update(SILVER_SESSION_TABLES)
-        if gold:
-            tables_to_generate.update(GOLD_SESSION_TABLES)
-        for table_name in tables_to_generate:
-            required_data.update(TABLE_REQUIREMENTS[table_name])
+            self._load_default_silver_tables()
+            silver_tables_to_generate = [self.data_lake.get("silver", table_name) for table_name, info in self.data_lake.metadata.items() if info["table_type"] == "silver"]
+            tables_to_generate.update(silver_tables_to_generate)
+            # refine sources for each silver table
+            for silver_table in silver_tables_to_generate:
+                silver_table.refine_sources()
+                required_data.update(set(silver_table.source_tables["bronze"]))
         
+        if gold:
+            gold_tables_to_generate = [self.data_lake.get("gold", table_name) for table_name, info in self.data_lake.metadata.items() if info["table_type"] == "gold"]
+            tables_to_generate.update(gold_tables_to_generate)
+            # refine sources for each gold table
+            for gold_table in gold_tables_to_generate:
+                gold_table.refine_sources()
+                required_data.update(set(gold_table.source_tables["bronze"]))
+
         # Use the unified get_data method instead of get_data_parallel
-        self.get_data(list(required_data), parallel=True)
+        logger.info(f"Topics to be loaded : {list(required_data)}")
+        self.get_data(list(required_data), parallel=False)
 
         self.first_datetime = self._get_first_datetime()
-        # self.session_start_time = self._get_session_start_time()
         self.session_start_datetime = self._get_session_start_datetime()
-        
-        if silver:
-            logger.info(f"Silver tables are being generated.")
-            for table_name in SILVER_SESSION_TABLES:
-                if table_name in TABLE_GENERATION_FUNCTIONS:
-                    setattr(self, table_name, self.data_lake.silver_lake.generate_table(table_name))
-                    logger.info(f"'{table_name}' has been generated and saved to the silver lake. You can access it from 'session.{table_name}'.")
-        
-        if gold:
-            logger.info("Gold tables are not implemented yet.")
-            pass
-    
-    def generate_laps_table(self):
-        setattr(self, "laps", self.data_lake.silver_lake.generate_table("laps"))
-    
-    def generate_car_telemetry_table(self):
-        setattr(self, "car_telemetry", self.data_lake.silver_lake.generate_table("laps"))
 
+        if self.data_lake._check_circular_dependencies():
+            if silver:
+                logger.info(f"Silver tables are being generated.")
+                for silver_table in silver_tables_to_generate:
+                    table_name = silver_table.table_name
+                    silver_table.generate_table()
+                    setattr(self, table_name, self.get_data(dataNames = table_name, level = "silver"))
+                    logger.info(f"'{table_name}' has been generated and saved to the silver lake. You can access it from 'session.{table_name}'.")
+            
+            if gold:
+                logger.info("Gold tables are being generated.")
+                for gold_table in gold_tables_to_generate:
+                    table_name = gold_table.table_name
+                    gold_table.generate_table()
+                    setattr(self, table_name, self.get_data(dataNames = table_name, level = "gold"))
+                    logger.info(f"'{table_name}' has been generated and saved to the gold lake. You can access it from 'session.{table_name}'.")
+        else:
+            logger.error("Circular dependencies detected. Please check your table dependencies.")
+
+
+    def _create_table(self, level, table_name, source_tables, include_session=False):
+        """
+        Decorator factory that creates a SilverTable instance
+        
+        Args:
+            table_name: Name for the silver table
+            source_tables: List of source table names to use from bronze lake
+        
+        Returns:
+            Decorator function that will register the callback for table generation
+        """
+        import inspect
+
+        if include_session:
+            param_table = ["_session"] + source_tables
+        
+       
+        if level == "silver":
+            source_dict = {
+                "bronze" : [],
+                "silver" : [],
+                "unknown" : []
+            }
+
+        elif level == "gold":
+            source_dict = {
+                "bronze" : [],
+                "silver" : [],
+                "gold" : [],
+                "unknown" : []
+            }
+        
+        def decorator(callback_func):
+
+            # Inspect the function signature to get parameter names
+            sig = inspect.signature(callback_func)
+            param_names = list(sig.parameters.keys())
+            
+            # Create mapping from source tables to parameter names
+            if len(param_names) != len(param_table):
+                raise ValueError(f"Function {callback_func.__name__} has {len(param_names)} parameters, but {len(param_table)} source tables were specified.\nRequired parameters: {param_table}")
+            param_mapping = dict(zip(param_table, param_names))
+
+            # # Check if the table is a topic name and if it is, add it to the source_dict
+            # for table, param_name in param_mapping.items():
+            #     if table == "_session": continue
+            #     else:
+            #         table_level = self._identify_data_level(table)
+            #         print(table, "-", table_level)
+            #         if table_level: source_dict[table_level].append(table)
+            #         else:
+            #             # source_dict["unknown"].append(table)
+            #             # logger.info(f"The table name you provided '{table}' is not loaded yet, it is expected to be loaded in the future. PLEASE MAKE SURE IT IS GOING TO BE CREATED.")
+            #             raise TopicNotFoundError(f"The table name you provided '{table}' is not a topic name.\nIf it is a table name from another level, SOURCE TABLES FROM ANOTHER LEVELS ARE NOT SUPPORTED YET.\nPlease call the table by `session.get_data('{table}')` inside the callback you are decorating.")
+
+            # source_objs = [] # TODO: Add sources from different levels
+
+            @functools.wraps(callback_func)
+            def wrapped_callback(*args, **kwargs):
+                # Get the source tables from bronze lake
+                source_data = {}
+                for table, param_name in param_mapping.items():
+                    if table == "_session": 
+                        source_data[param_name] = self
+                    else:
+                        # TODO: Add sources from different levels
+                        table_level = self.data_lake._identify_table_level(table)
+                        # source_dict[table_level].append(table)
+                        if table_level == None: raise TopicNotFoundError(f"The topic name you provided '{table}' is not included in any level.")
+                        elif table_level == "bronze":
+                            topic_name = self.check_data_name(table)
+                        else: topic_name = table
+
+                        # table_level = "bronze"
+                        # topic_name = self.check_data_name(table)
+
+                        temp_table = self.get_table(topic_name, level=table_level)
+                        source_data[param_name] = temp_table.df
+                        # source_objs.append(temp_table) # TODO: Add sources from different levels
+
+                # Call the original function with source data as parameters
+                result = callback_func(**source_data)
+                return result
+            
+            # if include_session:
+            #     source_dict["bronze"].append("_session")
+
+            if level == "silver":
+                # Create a new SilverTable instance
+                new_table = SilverTable(table_name=table_name, sources=source_tables)
+                new_table.callback = wrapped_callback
+            
+            elif level == "gold":
+                # Create a new SilverTable instance
+                new_table = GoldTable(table_name=table_name, sources=source_tables)
+                new_table.callback = wrapped_callback
+
+            # Store the table in the selected lake
+            self.data_lake.put(level=level, table_name=table_name, table=new_table)
+            
+            # Return the original function for documentation purposes
+            return callback_func
+        
+        logger.info(f"The callback function for the SILVER table '{table_name}' was set.")
+
+        return decorator
+
+    def create_silver_table(self, table_name, source_tables, include_session=False):
+
+        return self._create_table(
+            level = "silver",
+            table_name = table_name,
+            source_tables = source_tables,
+            include_session = include_session
+            )
+        
+    def create_gold_table(self, table_name, source_tables, include_session=False):
+        
+        return self._create_table(
+            level = "gold",
+            table_name = table_name,
+            source_tables = source_tables,
+            include_session = include_session
+            )
 
 def load_single_data(dataName, session, stream):
 
     if stream: dataType = "StreamPath"
     else: dataType = "KeyFramePath"
 
-    logger.info(f"Fetching data : '{dataName}'")
+    logger.debug(f"Fetching data : '{dataName}'")
     start = time()
     data = livetimingF1_getdata(
         urljoin(session.full_path, session.topic_names_info[dataName][dataType]),
@@ -598,45 +820,9 @@ def load_single_data(dataName, session, stream):
     logger.debug(f"Fetched in {round(time() - start,3)} seconds")
     # Parse the retrieved data using the ETL parser and return the result.
     start = time()
-    res = BasicResult(
-        data=list(session.etl_parser.unified_parse(dataName, data))
-    )
+    parsed_data = list(session.etl_parser.unified_parse(dataName, data))
+
     logger.debug(f"Parsed in {round(time() - start,3)} seconds")
     logger.info(f"'{dataName}' has been fetched and parsed")
 
-    # session.data_lake.put(
-    #     level="bronze", data_name=dataName, data=res
-    # )
-    # logger.debug(f"'{dataName}' has been saved to the bronze lake.")
-
-    return dataName, res
-
-
-# session.load()
-# session.generate(silver=True, gold=False)
-
-# session.load(
-#     bronze=True,
-#     silver=True,
-#     gold=True
-#     )
-
-# session.telemetry
-# session.timing
-# session.weather
-# session.position
-
-# telemetry
-# coordinates
-# tyre
-# stint
-# position
-
-# laps
-# pitduration
-# pitstops
-# timings
-
-# bronzeLake
-# silverLake
-# goldLake
\ No newline at end of file
+    return dataName, data, parsed_data
\ No newline at end of file
